//
// SEQUENCE.java --
//
// Author: Pekka Nikander <pnr@tequila.nixu.fi>
//
// Copyright 1996 by Nixu Oy, Helsinki, Finland
//
// Created: Wed Mar 13 17:18:43 1996
// Last modified: Sun Mar 31 22:51:54 1996 pnr
//
// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.asn1;

import cryptiX.utils.ArrayEnumeration;
import cryptiX.utils.Composite;

/**
 * This class represents ASN.1 sequences, and application
 * specific types derived from sequences.  Note that sequences
 * are NOT public.  They shall be created with ASN.create(ASN.SEQUENCE),
 * and accessed through the public OBJECT interface.
 */

final class SEQUENCE extends ASN1Object {
  /*
   * Initialize the object factory method.
   */
  static {
    Class _SEQUENCE = new SEQUENCE().getClass();
    ASN.register(ASN.SEQUENCE, _SEQUENCE);
    ASN.register(ASN.SET, _SEQUENCE);
  }

  /**
   * Default initial size of a sequence.
   * In SNMP applications this is 2, since most SEQUENCE
   * are just two elements long.
   */
  private final static int INITIAL_SIZE = 2;

  /**
   * Create a new SEQUENCE.  Called by the OBJECTFactory.
   */
  protected SEQUENCE() {
    contentData = new ASN1Object[INITIAL_SIZE];
  }

  /**
   * Perform SEQUENCE specific initialization
   */
  protected void init(ASN type) {
    this.asnType = type;
  }

  /******************************************************************
   * cryptiX.utils.Composite interface
   *****************************************************************/

  /**
   * The buffer where the contents are stored.
   */
  protected ASN1Object[] contentData;

  /**
   * The number of objects in the buffer.
   */
  protected int contentCount;

  /**
   * Add a new component to the end of sequence.
   */
  public void addComponent(Composite component) {
    addComponent(component, contentCount);
  }

  /**
   * Add a new component to a given location within the sequence.
   * @param component the component to insert
   * @param index where to insert the new component
   * @exception ArrayIndexOutOfBoundsException If the index is invalid.
   */
  public synchronized void addComponent(Composite component, int index) {
    if (contentCount >= contentData.length) {
      ASN1Object[] oldData = contentData;
      contentData = new ASN1Object[oldData.length * 2];
      System.arraycopy(oldData, 0, contentData, 0, oldData.length);
    }
    if (index != contentCount) {
      System.arraycopy(contentData, index, contentData, index + 1,
		       contentCount - index);
    }
    contentData[index] = (ASN1Object)component;
    contentCount++;
  }

  /**
   * Removes a component from a sequence.  If the component
   * is not within the sequence, does nothing.
   * @param component the component to remove
   */
  public synchronized void removeComponent(Composite component) {
    for (int i = 0; i < contentCount; i++) {
      if (component.equals(contentData[i])) {
	removeComponent(i);
	return;
      }
    }
  }

  /**
   * Removes a component at the given location within a sequence.
   * @param index where to remove the element from
   * @exception ArrayIndexOutOfBoundsException If the index is invalid.
   */
  public synchronized void removeComponent(int index) {
    if (index > contentCount)
      throw new ArrayIndexOutOfBoundsException();
    contentData[index] = null;
    if (contentCount - index - 1 > 0) {
      System.arraycopy(contentData, index + 1, contentData, index,
		       contentCount - index - 1);
    }
    contentCount--;
    contentData[contentCount] = null;
  }

  /**
   * Returns the component at the given location within a sequence.
   * @exception ArrayIndexOutOfBounds
   */
  public cryptiX.utils.Composite getComponent(int index) {
    if (index > contentCount)
      throw new ArrayIndexOutOfBoundsException();
    return contentData[index];
  }

  /**
   * Sets the component at the given location within a sequence.
   * @exception ArrayIndexOutOfBounds
   */
  public void setComponent(int index, cryptiX.utils.Composite value) {
    if (index > contentCount)
      throw new ArrayIndexOutOfBoundsException();
    contentData[index] = (ASN1Object)value;
  }

  /**
   * Returns the number of component objects.  Zero if none.
   */
  public int countComponents() {
    return contentCount;
  }

  /**
   * Returns an Enumeration that iterates over the component objects.
   * Returns a NullEnumeration if this object is not a container.
   */
  public synchronized java.util.Enumeration getComponents() {
    return new ArrayEnumeration(contentData, contentCount);
  }
  /****************************************************************
   * java.lang.Object interface
   ****************************************************************/

  /**
   * Returns a String that represents the value of this Object.
   */
  public String toString() {
    return asnType.getName()+" = "+"["+getClass().getName()+"] = " + contentCount + " elements";
  }
}
