// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.rsa;

import java.io.*;
import java.util.Random;

import cryptiX.asn1.*;
import cryptiX.crypto.*;
import cryptiX.utils.Util;

class RSAFunctions
{
    /**
     * mathematical functions for the RSA algorithm
     */

    /**
     * Modulo Multiplication: calculates a*b (mod n)
     */
    public static BigInt mMul(BigInt a, BigInt b, BigInt n) {

        BigInt a1 = a.mod(n);
        BigInt b1 = b.mod(n);

        BigInt product = a1.multiply(b1);

        return product.mod(n);
    }

    /**
     * Modulo Exponentiation: calculates a^b (mod n)
     */
    public static BigInt mExp(BigInt a, BigInt b, BigInt n) {

        BigInt r;
        BigInt v = new BigInt(a);
        BigInt p = new BigInt(b);
        BigInt tmp;

        if (p.testLSB())
            r = a;
        else
            r = BigInt.ONE;

        // while (p.compare(BigInt.ZERO) == 1) {
        while (!p.isZero()) {

            tmp = BigInt.multiply(v, v);
            v = tmp.mod(n);

            p.shiftRight();
            if (p.testLSB()) {
                tmp = BigInt.multiply(r, v);
                r = tmp.mod(n);
            }
        }

        return r;
    }


    /**
     * Multiplicative Inverse: solves ax = 1 (mod n)
     *
     * @param a variable a
     * @param n the modulus
     * @return variable x
     */
    public static BigInt mulInv(BigInt a, BigInt n) {

        BigInt z, y, g, gg, u, uu, v, vv;

        g = new BigInt(n); gg = new BigInt(a);
        u = BigInt.ONE;    v = BigInt.ZERO;
        uu = BigInt.ZERO;  vv = BigInt.ONE;

        while (!gg.isZero()) {
            y = BigInt.quotient(g, gg);

            z = g.subtract(y.multiply(gg)); g = gg; gg = z;
            z = u.subtract(y.multiply(uu)); u = uu; uu = z;
            z = v.subtract(y.multiply(vv)); v = vv; vv = z;
        }

        if (g.compare(BigInt.ONE) != 0) {
            System.out.println("a and n are not relativ prim!");
            System.out.println("multiplicative inverse doesn't exist!");
            return BigInt.ZERO;
        }
        if (v.test() == 1)      // positive
            return v;
        else
            return v.add(n);
    }

    /**
     * Binary gcd algorithm called by the public gcd
     */
    private static BigInt gcdBin(BigInt a, BigInt b) {
        BigInt t;
        int shifts = 0;

        while (true) {
            if (b.isZero()) break;

            if (a.testLSB()) {          // a is odd
                if (b.testLSB()) {      // b is odd
                    a = a.subtract(b);
                    a.shiftRight();
                    if (a.lessThan(b)) {
                        t = a; a = b; b = t;
                    }
                }
                else {                  // a is odd, b is even
                    b.shiftRight();
                    if (a.lessThan(b)) {
                        t = a; a = b; b = t;
                    }
                }
            }
            else {                      // a is even
                if (b.testLSB()) {      // b is odd
                    a.shiftRight();
                    if (a.lessThan(b)) {
                        t = a; a = b; b = t;
                    }
                }
                else {                  // a = even, b = even
                    a.shiftRight();
                    b.shiftRight();
                    shifts++;
                }
            }
        }

        if (shifts > 0)
            a.shiftLeft(shifts);

        return a;
    }

    /**
     * calculates the greatest common divisor
     *
     * @param a b the two numbers
     * @return greatest common divisor
     */
    public static BigInt gcd(BigInt a, BigInt b) {

        BigInt a1 = new BigInt(a);
        BigInt b1 = new BigInt(b);
        BigInt t;

        if (a1.lessThan(b1)) {
            t = a1;
            a1 = b1;
            b1 = t;
        }

        t = gcdBin(a1, b1);

        return t;

    }

    /**
     * functions for handling ASN1 types defined by RSA Laboratories
     **/

    public static String rsaEncryption =         new String("1.2.840.113549.1.1.1");
    public static String md2WithRSAEncryption =  new String("1.2.840.113549.1.1.2");
    public static String md4WithRSAEncryption =  new String("1.2.840.113549.1.1.3");
    public static String md5WithRSAEncryption =  new String("1.2.840.113549.1.1.4");
    public static String sha1WithRSAEncryption = new String("1.2.840.113549.1.1.5");

    public static String pbeWithMD5AndDES_CBC =  new String("1.2.840.113549.1.5.3");

    public static String rc4 = new String("1.2.840.113549.3.4");

    public static String md2 = new String("1.2.840.113549.2.2");
    public static String md4 = new String("1.2.840.113549.2.4");
    public static String md5 = new String("1.2.840.113549.2.5");
    public static String sha1 = new String("1.3.14.3.2.26");

    static {
        ASN1.registerOID("1.2.840.113549.1.1.1", "rsaEncryption");
        ASN1.registerOID("1.2.840.113549.1.1.2", "md2WithRSAEncryption");
        ASN1.registerOID("1.2.840.113549.1.1.3", "md4WithRSAEncryption");
        ASN1.registerOID("1.2.840.113549.1.1.4", "md5WithRSAEncryption");
        ASN1.registerOID("1.2.840.113549.1.1.5", "sha1WithRSAEncryption");
        ASN1.registerOID("1.2.840.113549.1.5.3", "pbeWithMD5AndDES_CBC");
        ASN1.registerOID("1.2.840.113549.3.4",   "rc4");
        ASN1.registerOID("1.2.840.113549.2.2",   "md2");
        ASN1.registerOID("1.2.840.113549.2.4",   "md4");
        ASN1.registerOID("1.2.840.113549.2.5",   "md5");
        ASN1.registerOID("1.3.14.3.2.26",        "sha1");
    }

    /**
     * this function parses a PKCS#1 defined encryption block
     * format: EB = BT|PS|00|D
     *
     * @param blockType the block type this block should have
     * @param encryptionBlock the block to be parsed
     **/
    public static byte[] parseEncryptionBlock(int blockType, byte[] encryptionBlock)
        throws CodingException {

        int len = encryptionBlock.length;
        int index = 0;                                      // index in encryption block
        if (encryptionBlock[index] != (byte)blockType) {    // check the block type
            throw new CodingException("Encryption block has wrong block type!");
        }

        // @todo check syntax according to block type

        while (encryptionBlock[++index] != 0)       // data starts after byte with value 0
            if (index >= len) {
                throw new CodingException("Encryption block has wrong format!");
            }

        if (index-1 < 8)
            throw new CodingException("Parsing Error: Padding String must be at least 8 octets long!");

        byte[] data = new byte[len-index-1];
        System.arraycopy(encryptionBlock, index+1, data, 0, len-index-1);

        return data;
    }

    /**
     * this function creates a PKCS#1 defined encryption block
     * format: EB = 00|BT|PS|00|D
     *
     * @param blockType the block type of this block
     * @param data the data for the encryption block
     * @param k length of the modulus in octets
     **/
    public static byte[] createEncryptionBlock(int blockType, byte[] data, int k)
        throws CodingException {

        int len = data.length;
        int index = 0;
        byte ps = 0;
        int psLength = k-3-len;

                    // according to PKCS#1 the PS must be at least 8 octets long
        if ((blockType == 1 || blockType == 2) && psLength < 8)
            throw new CodingException("Padding String must be at least 8 octets long!");
                    // according to PKCS#1 the Modulus must be at least 328 bits long
        if (k<41)
            throw new CodingException("Modulus must be at least 328 bits!");

        byte[] encryptionBlock = new byte[k];

        if (blockType == 0) {               // set padding bytes according to block type
            ps = (byte)0x00;
            if (data[0] == 0)
                throw new CodingException("For block type 0 data must begin with nonzero value");
        }
        else if (blockType == 1)
            ps = (byte)0xff;

        encryptionBlock[index++] = 0;                           // first 0
        encryptionBlock[index++] = (byte)blockType;             // block type

        if (blockType == 2) {
            Random rg = new Random();
            for (int i=0; i<psLength; i++) {
                while ((ps = (byte)rg.nextInt()) == 0)          // nonzero padding values
                    ;
                encryptionBlock[index++] = ps;                  // random padding
            }
        }
        else {
            for (int i=0; i<psLength; i++)                      // padding string
                encryptionBlock[index++] = ps;
        }

        encryptionBlock[index++] = 0;                           // second 0

        System.arraycopy(data, 0, encryptionBlock, index, len); // the data

        return encryptionBlock;
    }


    /**
     * checks the signature of a message defined in PKCS#1
     *
     * @param message the message the signature was made of
     * @param signature the RSA encrypted DigestInfo ASN1 type
     * @param pk the public key to verify the signature
     * @return true if signature is OK
     */
    public static boolean checkSignature(byte[] message, byte[] signature, PublicKey pk)
        throws CodingException {

        String digestAlgorithmID = null;
        byte[] decryptedDigest;
        byte[] digest;
                                    // 1: bit-string-to-octet-string conversion
                                    //      already done by DER decoding function
                                    // 2: RSA decryption
                                    // 2.1: octet-string-to-integer conversion
                                    // 2.2: RSA computation
        BigInt decryptSig = pk.crypt(new BigInt(signature));
                                    // 2.3: integer-to-octet-string conversion
                                    // 2.4: encryption block parsing
        try {
            byte[] data = parseEncryptionBlock(0x01, decryptSig.toByteArray());

            if (data == null)           // if the encryption blockb has the wrong
                return false;           // format the public key or the signature is wrong

                                        // 3: Data decoding, shall be a DigestInfo
            ASN1 digestInfo = new ASN1(data);

            digestAlgorithmID = Util.AlgIDAlgorithm(digestInfo.getComponentAt(0));
            decryptedDigest = (byte[])(digestInfo.getComponentAt(1)).getValue();
        } catch (CodingException ex) {
            return false;
        }

        if (digestAlgorithmID.equals(md2)) {
            throw new CodingException("MD2 not implemented!");
        }
        else if (digestAlgorithmID.equals(md4)) {
            throw new CodingException("MD4 not implemented!");
        }
        else if (digestAlgorithmID.equals(md5)) {
                                    // 4: Message digesting and comparison
            Md5Hash hash = new Md5Hash();
            hash.add(message);
            digest = hash.get();

            for (int i=0; i<hash.hashSize(); i++)
                if (digest[i] != decryptedDigest[i])
                    return false;

            return true;
        }
        else if (digestAlgorithmID.equals(sha1)) {
                                    // 4: Message digesting and comparison
            ShaHash hash = new ShaHash();
            hash.add(message);
            digest = hash.get();

            for (int i=0; i<hash.hashSize(); i++)
                if (digest[i] != decryptedDigest[i])
                    return false;

            return true;
        }
        else
            throw new CodingException("Hash Algorithm not implemented!");
    }

    /**
     * creates a signature from a message
     *
     * @param message the data which should be signed
     * @param sk the private key to sign the message
     * @param digestAlgorithmID ID of digest algorithm
     * @return the signature
     */
    public static byte[] createSignature(byte[] message, PrivateKey sk, String digestAlgorithmID)
        throws CodingException {

        byte[] digest = null;

        if (digestAlgorithmID.equals(md2)) {
            throw new CodingException("MD2 not implemented!");
        }
        else if (digestAlgorithmID.equals(md4)) {
            throw new CodingException("MD4 not implemented!");
        }
        else if (digestAlgorithmID.equals(md5)) {
                                    // 1: Message digesting
            Md5Hash md5 = new Md5Hash();
            md5.add(message);
            digest = md5.get();
        }
        else if (digestAlgorithmID.equals(sha1)) {
                                    // 1: Message digesting
            ShaHash sha = new ShaHash();
            sha.add(message);
            digest = sha.get();
        }
        else
            throw new CodingException("Hash Algorithm not implemented!");

                                    // 2: Data encoding (DigestInfo)
        ASN1Object di = null;

        try {
            di = ASN.create(ASN.SEQUENCE);
            di.addComponent(Util.AlgID(digestAlgorithmID));
            di.addComponent(ASN.create(ASN.OCTET_STRING, digest));
        } catch (InstantiationException ex) {
        }

        int k = (sk.modulus().bitLength()-1) / 8 + 1;        // length of the modulus in octets

        byte[] data = createEncryptionBlock(0x1, new ASN1(di).toByteArrayDER(), k);
                                    // 3: RSA encryption
        BigInt encryptSig = sk.cryptPrivate(new BigInt(data));

        return encryptSig.toByteArray();
                                    // 4: Octet-string-to-bit-string conversion
                                    //    is done by calling method
    }
}
