// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.rsa;

import java.io.IOException;
import cryptiX.asn1.*;
import cryptiX.utils.*;
import cryptiX.crypto.*;



class DES_CBCkeyGen {

    byte[] desKey = new byte[8];
    byte[] iv = new byte[8];

    public DES_CBCkeyGen(String password, byte[] saltValue, int iterationCount) {

                // now generate the DES KEY and the initialisation vector for CBC mode
        Md5Hash hash = new Md5Hash();

        for (int i=0; i<iterationCount; i++) {      // iterationCount times the hash on P|S
            hash.add(password);
            hash.add(saltValue);
        }

        byte[] tmp = hash.get();

        System.arraycopy(tmp, 0, desKey, 0, 8);     // the first 8 bytes build DES key
        System.arraycopy(tmp, 8, iv, 0, 8);         // the second 8 bytes build IV

        for (int i=0; i<8; i++)
            if ( Util.even(Util.countOnes(desKey[i])))  // check odd parity for DES key octets
		        desKey[i] ^= 0x1;                       // otherwise change LSB
    }

    public byte[] DesKey() {
        return desKey;
    }

    public byte[] Iv() {
        return iv;
    }
}


public class RSAPrivateKey extends PrivateKey {

    private ASN1 privateKey = null;                 // the private key in any form
    private ASN1 rsaPrivateKey = null;              // RsaPrivateKey
    private ASN1 privateKeyInfo = null;             // PrivateKeyInfo
    private ASN1 encryptedPrivateKeyInfo = null;    // encrypted PrivateKeyInfo

    private int saveFormat = ASN1.DER;      // default save format is DER
    private String encAlgId = null;         // the algorithm used to encrypt the private key

    private BigInt RPKversion = BigInt.ZERO;        // version of RSAPrivateKey
    private BigInt PKIversion = BigInt.ZERO;        // version of PrivateKeyInfo
    private boolean encrypted = false;              // the private key is not encrypted
    private boolean netscapeFormat = false;         // true, if the key is in netscape format

    /**
     * Default constructor.
     */
    public RSAPrivateKey() {
        super();

        new ASN1();         // used to load the classes for generating ASN.1 objects
    }

    /**
     * creates a RSA private key from a PEM of  DER file
     */
    public RSAPrivateKey(String fileName) throws IOException, CodingException  {
        privateKey = new ASN1(fileName);
        saveFormat = privateKey.saveFormat();
        init();
    }

    /**
     * creates a RSA private key from an ASN1Object
     */
    public RSAPrivateKey(ASN1Object o) throws CodingException {
        privateKey = new ASN1(o);
        init();
    }

    /**
     * generates the key pair
     *
     * @param keylen number of bits for modulus
     */
    public void generateKeys(int keylen) {

        super.generateKeys(keylen);
    }

    /**
     * generates the key pair
     *
     * @param keylen number of bits for modulus
     * @param rng the random number generator to use
     */
    public void generateKeys(int keylen, java.util.Random rng) {

        super.generateKeys(keylen, rng);
    }

    /**
     * initialize a private key from an object of the following type
     *      - RSAPrivateKey
     *      - PrivateKeyInfo
     *      - EncryptedPrivateKeyInfo
     *      - EncryptedPrivateKeyInfo in netscape format
     */
    private void init() throws CodingException {

        if (privateKey.components() == 2) {

            ASN1Object obj = privateKey.getComponentAt(0);

            if (obj.isA(ASN.OCTET_STRING)) {            // Netscape format
                netscapeFormat = true;
                encryptedPrivateKeyInfo = new ASN1(privateKey.getComponentAt(1));
            }
            else
                encryptedPrivateKeyInfo = privateKey;

                        // get the AlgorithmIdentifier
            encAlgId = Util.AlgIDAlgorithm(encryptedPrivateKeyInfo.getComponentAt(0));
            encrypted = true;
        }
        else if (privateKey.getComponentAt(1).isA(ASN.SEQUENCE)) {
            privateKeyInfo = privateKey;
            parsePrivateKeyInfo();
        }
        else {
            rsaPrivateKey = privateKey;
            parseRsaPrivateKey();
        }
    }

    /**
     * decrypts an encrypted PrivateKeyInfo (PKCS#5 and PKCS#8).
     *
     * @param password the password to decrypt the key
     */
    public void decrypt(String password) throws CodingException {

        byte[] message;     // the message in the encryption block
        byte[] eb = null;   // the encryption block

                    // get the encrypted data
        byte[] encryptedData = (byte[])encryptedPrivateKeyInfo.getComponentAt(1).getValue();

        if (encAlgId.equals(RSAFunctions.pbeWithMD5AndDES_CBC)) {
                        // get the parameters of the algorithm
            ASN1 pbeParameter = new ASN1(Util.AlgIDParameter(encryptedPrivateKeyInfo.getComponentAt(0)));
                        // get salt value from parameters
            byte[] saltValue = (byte[])pbeParameter.getComponentAt(0).getValue();
                        // get iteration from parameters
            int iterationCount = ((BigInt)pbeParameter.getComponentAt(1).getValue()).intValue();
                        // generate the DES key and the IV
            DES_CBCkeyGen kg = new DES_CBCkeyGen(password, saltValue, iterationCount);
                        // create a DES-CBC cipher
            CbcBlockCipher des_cbc = new CbcBlockCipher(new DesCipher(kg.DesKey()));
            des_cbc.setIv(kg.Iv());                         // set IV
            eb = des_cbc.decrypt(encryptedData);            // decrypt the private key info
            message = parseEncryptionBlock(eb);             // parse the encryption block
        }
        else if (encAlgId.equals(RSAFunctions.rc4)) {       // netscape rc4 encrypted

            Md5Hash hash = new Md5Hash();
            hash.addASCII(password);                        // MD5 hash of the password
            Rc4Cipher rc4 = new Rc4Cipher(hash.get());      // is used for the key

            message = rc4.decrypt(encryptedData);           // decrypt the private key info
        }
        else
            throw new CodingException("Algorithm "+encAlgId+" not implemented!");


        privateKeyInfo = new ASN1(message);
        parsePrivateKeyInfo();
    }

    /**
     * Parses a PrivateKeyInfo structure (PKCS#8).
     */
    private void parsePrivateKeyInfo() throws CodingException {

        String algId;

        try {
            PKIversion = (BigInt)(privateKeyInfo.getComponentAt(0)).getValue();

            algId = Util.AlgIDAlgorithm(privateKeyInfo.getComponentAt(1));

            if (algId.equals(RSAFunctions.rsaEncryption))
                 rsaPrivateKey = new ASN1((byte[])(privateKeyInfo.getComponentAt(2)).getValue());
            else
                throw new CodingException("Algorithm "+algId+" not implemented!");

        } catch (ClassCastException ex) {
            throw new CodingException("No RSA Private Key!");
        }

        parseRsaPrivateKey();
    }

    /**
     * Parses a RSAPrivateKey structure according to PKCS#1.
     */
    private void parseRsaPrivateKey() throws CodingException{

        try {
            RPKversion =        (BigInt)(rsaPrivateKey.getComponentAt(0)).getValue();
            modulus =           (BigInt)(rsaPrivateKey.getComponentAt(1)).getValue();
            publicExponent =    (BigInt)(rsaPrivateKey.getComponentAt(2)).getValue();
            privateExponent =   (BigInt)(rsaPrivateKey.getComponentAt(3)).getValue();
            prime1 =            (BigInt)(rsaPrivateKey.getComponentAt(4)).getValue();
            prime2 =            (BigInt)(rsaPrivateKey.getComponentAt(5)).getValue();
            exponent1 =         (BigInt)(rsaPrivateKey.getComponentAt(6)).getValue();
            exponent2 =         (BigInt)(rsaPrivateKey.getComponentAt(7)).getValue();
            coefficient =       (BigInt)(rsaPrivateKey.getComponentAt(8)).getValue();
        } catch (ClassCastException ex) {
            throw new CodingException("No RSA Private Key!");
        }
    }

    /**
     * Parses the encryption block after DES decrypting (PKCS#5).
     *
     * format: EB = M|PS    PS value = PS length = 8 - (|M| mod 8)
     *
     * @param eb the encryption block to parse
     * @return the original message
     */
    private byte[] parseEncryptionBlock(byte[] eb) throws CodingException {
        int length = eb.length;
        int paddingString = eb[length-1];               // the last value is always a PS

        if (paddingString < 0)
            throw new CodingException("Encryption block parsing error!");

        for (int i=1; i<=paddingString; i++)            // check the padding string
            if (eb[length-i] != paddingString)
                throw new CodingException("Encryption block parsing error!");

        byte[] message = new byte[length-paddingString];    // create the message
        System.arraycopy(eb, 0, message, 0, length-paddingString);

        return message;
    }

    /**
     * Create the encryption block before DES encrypting (PKCS#5).
     *
     * format: EB = M|PS    PS value = PS length = 8 - (|M| mod 8)
     *
     * @param message messsage to format
     * @return the formatted encryption block
     */
    private byte[] createEncryptionBlock(byte[] message) {

        int length = message.length;
        int paddingString = 8-(length % 8);             // length and value of PS

        if (paddingString == 0)                         // always use a PS
            paddingString = 8;

        byte[] eb = new byte[length+paddingString];     // the encryption block

        System.arraycopy(message, 0, eb, 0, length);    // the message

        for (int i=0; i<paddingString; i++)             // add PS
            eb[length+i] = (byte)paddingString;

        return eb;
    }


    /**
     * Encrypt the PrivateKeyInfo data structure (PKCS#5 and PKCS#8).
     *
     * @param password the password to use
     * @return EncryptedPrivateKeyInfo
     */
    public void encrypt(String password) throws CodingException {

        byte[] saltValue = new byte[8];
        byte[] encryptedData = null;
        int iterationCount = 3;

        if (privateKeyInfo == null)
            createPrivateKeyInfo();

        byte[] message = privateKeyInfo.toByteArrayDER();

        if (encAlgId.equals(RSAFunctions.pbeWithMD5AndDES_CBC)) {

                    // generate the salt value as a hash from P|M
            Md5Hash hash = new Md5Hash();
            hash.addASCII(password);
            hash.add(message);
            byte[] tmp = hash.get();
            System.arraycopy(tmp, 0, saltValue, 0, 8);       // copy the first 8 bytes

            DES_CBCkeyGen kg = new DES_CBCkeyGen(password, saltValue, iterationCount);

            byte[] eb = createEncryptionBlock(message);     // create the encryption block

            CbcBlockCipher des_cbc = new CbcBlockCipher(new DesCipher(kg.DesKey()));

            des_cbc.setIv(kg.Iv());                         // set IV

            encryptedData = des_cbc.encrypt(eb);            // encrypt the private key
        }
        else if (encAlgId.equals(RSAFunctions.rc4)) {

            Md5Hash hash = new Md5Hash();
            hash.addASCII(password);

            Rc4Cipher rc4 = new Rc4Cipher(hash.get());

            encryptedData = rc4.decrypt(message);           // decrypt the private key info
        }
        else
            throw new CodingException("Algorithm "+encAlgId+" not implemented!");

        ASN1Object epki = null;
        ASN1Object pbeParameter = null;

        try {
            epki = ASN.create(ASN.SEQUENCE);

            pbeParameter = ASN.create(ASN.SEQUENCE);
            pbeParameter.addComponent(ASN.create(ASN.OCTET_STRING, saltValue));
            pbeParameter.addComponent(ASN.create(ASN.INTEGER, new BigInt(iterationCount)));

            epki.addComponent(Util.AlgID(encAlgId, pbeParameter));
            epki.addComponent(ASN.create(ASN.OCTET_STRING, encryptedData));

        } catch (InstantiationException ex) {
        }

        encrypted = true;
        encryptedPrivateKeyInfo = new ASN1(epki);
        privateKey = encryptedPrivateKeyInfo;
    }

    /**
     * Creates a RSAPrivateKey data structure (PKCS#1).
     */
    private void createRsaPrivateKey() {
        ASN1Object rpk = null;

        try {
            rpk = ASN.create(ASN.SEQUENCE);

            rpk.addComponent(ASN.create(ASN.INTEGER, RPKversion));
            rpk.addComponent(ASN.create(ASN.INTEGER, modulus));
            rpk.addComponent(ASN.create(ASN.INTEGER, publicExponent));
            rpk.addComponent(ASN.create(ASN.INTEGER, privateExponent));
            rpk.addComponent(ASN.create(ASN.INTEGER, prime1));
            rpk.addComponent(ASN.create(ASN.INTEGER, prime2));
            rpk.addComponent(ASN.create(ASN.INTEGER, exponent1));
            rpk.addComponent(ASN.create(ASN.INTEGER, exponent2));
            rpk.addComponent(ASN.create(ASN.INTEGER, coefficient));
        } catch (InstantiationException ex) {
        }
        rsaPrivateKey = new ASN1(rpk);
        privateKey = rsaPrivateKey;
    }

    /**
     * Creates a PrivateKeyInfo data structure (PKCS#8).
     */
    private void createPrivateKeyInfo() {
        ASN1Object pki = null;

        if (rsaPrivateKey == null)
            createRsaPrivateKey();

        try {
            pki = ASN.create(ASN.SEQUENCE);

            pki.addComponent(ASN.create(ASN.INTEGER, PKIversion));
            pki.addComponent(Util.AlgID(RSAFunctions.rsaEncryption));
            pki.addComponent(ASN.create(ASN.OCTET_STRING, rsaPrivateKey.toByteArrayDER()));
        } catch (InstantiationException ex) {
        }
        privateKeyInfo = new ASN1(pki);
        privateKey = privateKeyInfo;
    }

    /**
     * Creates a private key in netscape format.
     */
    private void createNetscapePrivateKey() {

        ASN1Object npk = null;           // netscape private key

        if (privateKey == null)
            createPrivateKeyInfo();

        try {
            npk = ASN.create(ASN.SEQUENCE);

            byte[] tmp = new byte[11];
            new String("private-key").getBytes(0, 11, tmp, 0);

            npk.addComponent(ASN.create(ASN.OCTET_STRING, tmp));

            npk.addComponent(privateKey.toASN1Object());    // use the private key in any form
        } catch (InstantiationException ex) {
        }
        privateKey = new ASN1(npk);
    }

    /**
     * returns true if the private key is encrypted
     */
    public boolean isEncrypted() {
        return encrypted;
    }

    /**
     * Encrypt private key and save to file.
     *
     * @param fileName the file name to use
     * @param password the password for encrypting the private key
     * @param algorithmIdentifier algorithm to use for encrypting the key
     * @param pemFormat true  file is written in PEM format
     *                  false file is written in DER format
     **/
    public void saveToFile(String fileName) throws IOException {

        if (privateKey == null)
            createPrivateKeyInfo();

        if (netscapeFormat)
            createNetscapePrivateKey();

        if (saveFormat == ASN1.PEM)
            privateKey.saveToFile(fileName, "-----BEGIN RSA PRIVATE KEY-----", "-----END RSA PRIVATE KEY-----");
        else
            privateKey.saveToFile(fileName);
    }

    /**
     * returns the public key part of this private key
     */
    public PublicKey publicKey() {

        return new RSAPublicKey(publicExponent, modulus);
    }

    /**
     * sets the algorithm used to encrypt the private key
     */
    public void setEncryptionAlgorithm(String alg) {
        encAlgId = alg;
        privateKey = privateKeyInfo;    // algorithm has changed
        encrypted = false;              // key must be encrypted again
    }

    /**
     * sets the file format (ASN1.DER, ASN1.PEM) of the private key
     */
    public void setSaveFormat(int format) {
        saveFormat = format;
    }

    /**
     * sets this private key to netscape format
     */
    public void setNetscapeFormat(boolean nf) {
        netscapeFormat = nf;
        privateKey = privateKeyInfo;
    }

    /**
     * returns the algorithm the private key was encrypted
     */
    public String encryptionAlgorithm() {
        return encAlgId;
    }

    /**
     * returns the file format (ASN1.DER, ASN1.PEM) of the private key
     */
    public int saveFormat() {
        return saveFormat;
    }

    /**
     * returns true if this privateKey is in netscape format
     */
    public boolean netscapeFormat() {
        return netscapeFormat;
    }

    /**
     * returns the fingerprint of this private key
     */
    public byte[] fingerprint() {
        return privateKey.fingerprint();
    }
}

