// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.rsa;

import java.io.IOException;

import cryptiX.asn1.*;
import cryptiX.utils.Util;
import cryptiX.crypto.Md5Hash;

/**
 * This class is used to set the key values by parsing
 * a SubjectPublicKey ASN1 structure if the algorithm
 * identifier says rsaEncryption.
 */
public class RSAPublicKey extends PublicKey {

    private ASN1 subjectPublicKey;                  // SubjectPublicKey
    private ASN1 subjectPublicKeyInfo;              // SubjectPublicKeyInfo
    private String algorithmID = RSAFunctions.rsaEncryption;
    private int saveFormat = ASN1.DER;              // default save format is DER


    /**
     * creates a RSA public key from a PEM of  DER file
     */
    public RSAPublicKey(String fileName) throws IOException, CodingException  {
        ASN1 publicKey = new ASN1(fileName);
        saveFormat = publicKey.saveFormat();
        init(publicKey);
    }

    /**
     * creates a RSA public key with given values
     *
     * @param exponent the public exponent
     * @modulus the modulus
     */
    public RSAPublicKey(BigInt exponent, BigInt modulus) {
        this.modulus = modulus;
        publicExponent = exponent;
        ID = null;
    }

    /**
     * creates a public key from an ASN1Object of the following type
     *      - SubjectPublicKey
     *      - SubjectPublicKeyInfo
     *
     * @param publicKey the public key
     */
    public RSAPublicKey(ASN1Object publicKey) throws CodingException {

        init(new ASN1(publicKey));
    }

    /**
     * initialize a public key from an object of the following type
     *      - SubjectPublicKey
     *      - SubjectPublicKeyInfo
     */
    private void init(ASN1 publicKey) throws CodingException {

        ASN1Object obj = publicKey.getComponentAt(0);

        if (obj.isA(ASN.OID)) {                     // SubjectPublicKeyInfo
            subjectPublicKeyInfo = publicKey;
            algorithmID = Util.AlgIDAlgorithm(obj);

            subjectPublicKey = new ASN1((byte[])publicKey.getComponentAt(1).getValue());
        }
        else
            subjectPublicKey = publicKey;
                                                    // SubjectPublicKey
        modulus =           (BigInt)(subjectPublicKey.getComponentAt(0)).getValue();
        publicExponent =    (BigInt)(subjectPublicKey.getComponentAt(1)).getValue();

    }

    /**
     * creates a SubjectPublicKey object (X.509)
     */
    private void createSubjectPublicKey() {
        ASN1Object spk = null;

        try {
            spk = ASN.create(ASN.SEQUENCE);
            spk.addComponent(ASN.create(ASN.INTEGER, modulus));
            spk.addComponent(ASN.create(ASN.INTEGER, publicExponent));
        } catch (InstantiationException ex) {
        }

        subjectPublicKey = new ASN1(spk);
    }

    /**
     * creates a SubjectPublicKeyInfo object (X.509)
     */
    private void createSubjectPublicKeyInfo() {
        ASN1Object spki = null;

        if (subjectPublicKey == null)
            createSubjectPublicKey();

        try {
            spki = ASN.create(ASN.SEQUENCE);

            spki.addComponent(Util.AlgID(algorithmID));

                        // subject public key bit string
            byte[] spkbs = subjectPublicKey.toByteArrayDER();

            spki.addComponent(ASN.create(ASN.BIT_STRING, spkbs));
        } catch (InstantiationException ex) {
        }

        subjectPublicKeyInfo = new ASN1(spki);
    }

    /**
     * returns a SubjectPublicKeyInfo object
     * overwrites method in PublicKey
     */
    public ASN1Object toSubjectPublicKeyInfo() {
        createSubjectPublicKeyInfo();

        return subjectPublicKeyInfo.toASN1Object();
    }

    /**
     * returns a SubjectPublicKey object
     * overwrites method in PublicKey
     */
    public ASN1Object toSubjectPublicKey() {
        createSubjectPublicKey();

        return subjectPublicKey.toASN1Object();
    }
}

