// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.rsa;

import java.util.*;

/**
 * The generation of random and random prime numbers should be in BigInt class.
 * But BigInt is derived from Number and therefore could not extend Oberservable because
 * Java doesn't support multiple inheritance. If somebody knows a better solution let me know.
 */
public class RandomPrimeGenerator extends Observable {


    static int MASK = BigInt.getMask();
    static int BITS = BigInt.getBits();

    Random rng = null;

    /**
     * default constructor
     *
     * uses the Java default random number generator
     */
    RandomPrimeGenerator() {

        rng = new Random();
    }

    /**
     * uses a given random number generator
     */
    RandomPrimeGenerator(Random r) {

        if (r == null)
            rng = new Random();         // use default
        else
            rng = r;
    }


    /**
     * Generate a random BigInt less a specified limit
     * @param rng	the random number generator to use.
     *                  It is called an unspecified number of times.
     * @param limit	specifies the range of the result.
     * @return a BigInt in the range 0..limit-1 inclusive.
     */
    public BigInt random(BigInt limit) {

    	if (limit.negative || limit.isZero())
	      throw new ArithmeticException ("Random limit must be positive");
    	int[] digits = new int[limit.last+1];

	    while (true) {
	        // Generate most significant digit(s) carefully until one of the
    	    // digits is less than the corresponding digit in `limit'
	        int i = limit.last;
	        while (i>=0) {
    	    	int digit = limit.digits[i],  mask = 1,  trial;
	    	    while (mask<digit) mask=(mask<<1)+1;
    		    do {
	    	        trial = rng.nextInt() & mask;
    	    	} while (trial>digit);
	    	    digits[i--] = trial;
    		    if (trial<digit) {
	    	        // Now the number must be less than limit, so rest of the
    	    	    // digits can be anything.
	    	        while (i>=0)
		              digits[i--] = rng.nextInt() & MASK;
        		    return  new BigInt (false, digits, limit.last).trim();
	        	}
	        }
	    // We get here only if we generated a number equal to limit.
	    }
    }

    /**
     * Generate a random BigInt with specified length
     * @param rng	the random number generator to use.
     *                  It is called an unspecified number of times.
     * @param bitlen specifies the length of the result.
     * @return a BigInt with bitlen bits
     */
	public BigInt random(java.util.Random rng, int bitlen) {

    	int len = (bitlen-1)/BITS;      // last digit for storing the new BigInt
	    int[] digits = new int[len+1];
        int i=len;
        int re = bitlen-len*BITS;       // remainder of bits in last digit

        digits[i] = rng.nextInt() & ((1 << re)-1);
        digits[i--] |= (1 << (re-1));   // set msb
        // Now the number must be less than limit, so rest of the
        // digits can be anything.
        while (i>=0)
            digits[i--] = rng.nextInt() & MASK;

        return  new BigInt (false, digits, len).trim();
	}


    /**
     * Generate a prime BigInt.
     * @param rng	the random number generator to use.
     *                  It is called an unspecified number of times.
     * @param limit	specifies the range of the result.
     * @return a BigInt in the range 0..limit-1 inclusive.
     */
	public BigInt randomPrime(int bitlen) {

        BigInt rnd = random(rng, bitlen);

        if (!rnd.testLSB())             // number should be odd
            rnd = rnd.add(BigInt.ONE);

        while (!TestPrime.isPrime(rnd)) {

            setChanged();
            notifyObservers();
            rnd = rnd.add(BigInt.TWO);
        }

        return rnd;
    }
}
/*
    private static String animation[] = {"|", "/", "-", "\\"};
        int length = animation.length;
        j = j++ % length;
        int j = 0;

*/