// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.user;

import java.io.*;
import java.util.Date;
import cryptiX.asn1.*;
import cryptiX.rsa.*;
import cryptiX.crypto.Md5Hash;
import cryptiX.utils.*;

/**
 * This class represents a X509 certificate.
 */

public class X509Certificate {

    private ASN1 certificate = null;            // wether the certificate in netscape format
                                                // or the same as x509Certificate
    private ASN1 x509Certificate = null;        // certificate according to X.509
    private ASN1 certificateInfo = null;        // certificate info according to X.509

    private int saveFormat = ASN1.DER;          // default save format is DER
    private boolean netscapeFormat = false;     // is the cert in netscape format

    private int         version = 0;                    // version number
    private BigInt      serialNumber = BigInt.ZERO;     // serial number
    private String      signatureAlgorithmID = RSAFunctions.md5WithRSAEncryption;    // signature algorithm id
    private Name        issuer = null;                  // name of the issuer
    private Date        notBefore = null;               // validity not before
    private Date        notAfter = null;                // validity not after
    private Name        subject = null;                 // name of the subject
    private PublicKey   publicKey = null;               // the public inside this certificate
    private BigInt      issuerUniqueID = null;
    private BigInt      subjectUniqueID = null;
    private Extensions  v3Extensions = null;            // the extensions of version 3

    /**
     * create a x509Certificate form a PEM or DER file
     */
    public X509Certificate(String fileName) throws IOException, CodingException {
        certificate = new ASN1(fileName);
        saveFormat = certificate.saveFormat();

        if (certificate != null)
            init();
    }

    /**
     * create a x509Certificate form a PEM or DER array
     */
    public X509Certificate(byte[] arr) throws CodingException {
        certificate = new ASN1(arr);

        if (certificate != null)
            init();
    }

    /**
     * default constructor
     */
    public X509Certificate() {
        certificate = new ASN1();
        x509Certificate = certificate;
    }

    /**
     * inits the data from the x509 certificate
     */
    private void init() throws CodingException {

        int off = 0;         // offset if version is present
        ASN1Object obj;

        obj = certificate.getComponentAt(0);

        if (obj.isA(ASN.OCTET_STRING)) {        // if first component is a OCTET_STRING
            netscapeFormat = true;              // then this is a netscpae cert

            x509Certificate = new ASN1(certificate.getComponentAt(1));
        }
        else
            x509Certificate = certificate;

        certificateInfo = new ASN1(x509Certificate.getComponentAt(0));

        obj = certificateInfo.getComponentAt(0);

        if (obj.isA(ASN.CONTEXT_SPECIFIC)) {
            ASN1Object ver = (ASN1Object)obj.getValue();
            BigInt b = (BigInt)ver.getValue();          // cert contains version number
            version = b.intValue();                     // version = value + 1
            version++;
            off++;
        }

        serialNumber =          (BigInt)certificateInfo.getComponentAt(0+off).getValue();

        signatureAlgorithmID = Util.AlgIDAlgorithm(certificateInfo.getComponentAt(1+off));

        issuer = new Name(certificateInfo.getComponentAt(2+off));

        ASN1 valitity = new ASN1(certificateInfo.getComponentAt(3+off));
        notBefore = Util.UTCTime2Date((String)valitity.getComponentAt(0).getValue());
        notAfter = Util.UTCTime2Date((String)valitity.getComponentAt(1).getValue());


        subject = new Name(certificateInfo.getComponentAt(4+off));
        ASN1Object spki = certificateInfo.getComponentAt(5+off);
        publicKey = PublicKey.valueOf(spki);

        int index = 6+off;                          // index for parsing v2, v3 certs

        while (index < certificateInfo.components()) {

            CONTEXT contextObj = (CONTEXT)certificateInfo.getComponentAt(index);

            if (contextObj.tag() == 1)
                System.out.println("found issuerUniqueID");
            else if (contextObj.tag() == 2)
                System.out.println("found subjectUniqueID");
            else if (contextObj.tag() == 3)
                v3Extensions = new Extensions((ASN1Object)contextObj.getValue());

            index++;
        }
    }

    /**
     * creates the ASN1 object from the variables
     */
    private void createCertificateInfo() {

        ASN1Object ci = null;           // certificate info

        try {
            ci = ASN.create(ASN.SEQUENCE);

            if (version > 1) {
                ASN1Object ver = ASN.create(ASN.INTEGER, new BigInt((long)(version-1)));
                ci.addComponent(ASN.create(ASN.CONTEXT_SPECIFIC, new CONTEXT(0, ver)));
            }

            ci.addComponent(ASN.create(ASN.INTEGER, serialNumber));

            ci.addComponent(Util.AlgID(signatureAlgorithmID));

            ci.addComponent(issuer.toASN1Object());

            ASN1Object val = ASN.create(ASN.SEQUENCE);
            val.addComponent(ASN.create(ASN.UTCTIME, Util.Date2UTCTime(notBefore)));
            val.addComponent(ASN.create(ASN.UTCTIME, Util.Date2UTCTime(notAfter)));
            ci.addComponent(val);

            ci.addComponent(subject.toASN1Object());
            ci.addComponent(publicKey.toSubjectPublicKeyInfo());
        } catch (InstantiationException ex) {
        }
        certificateInfo = new ASN1(ci);
    }

    /**
     * creates a certificate in netscape format
     */
    private void createNetscapeCertificate() {

        ASN1Object nc = null;           // certificate info

        try {
            nc = ASN.create(ASN.SEQUENCE);

            byte[] tmp = new byte[11];
            new String("certificate").getBytes(0, 11, tmp, 0);

            nc.addComponent(ASN.create(ASN.OCTET_STRING, tmp));

            nc.addComponent(x509Certificate.toASN1Object());
        } catch (InstantiationException ex) {
        }
        certificate = new ASN1(nc);
    }


    /**
     * signs the certificate with the private key of the issuer
     *
     * @param issuerSK the private key of the issuer
     */
    public void signCertificate(PrivateKey issuerSK) throws CodingException {

        createCertificateInfo();

        byte[] signature = issuerSK.createSignature(certificateInfo.toByteArrayDER(), signatureAlgorithmID);

        ASN1Object c = null;

        try {
            c = ASN.create(ASN.SEQUENCE);
            c.addComponent(certificateInfo.toASN1Object());
            c.addComponent(Util.AlgID(signatureAlgorithmID));
            c.addComponent(ASN.create(ASN.BIT_STRING, signature));
        } catch (InstantiationException ex) {
        }

        x509Certificate = new ASN1(c);

        if (netscapeFormat)
            createNetscapeCertificate();
        else
            certificate = x509Certificate;
    }

    /**
     * verifies a self signed certificate
     *
     * @return true if signature OK
     */
    public boolean verify() throws CodingException {

        return verify(publicKey);
    }

    /**
     * verifies a signed certificate
     *
     * @param signerPK the public key from the signer
     * @return true if signature OK
     */
    public boolean verify(PublicKey signerPK) throws CodingException {

                        // get the signatureAlgorithm
        String signatureAlgorithm = Util.AlgIDAlgorithm(x509Certificate.getComponentAt(1));

                        // get the signature from the certificate
        ASN1Object sigObject = x509Certificate.getComponentAt(2);
        byte[] signature = (byte[])(sigObject).getValue();

                        // get the signed Data from the certificate
        byte[] signedData = x509Certificate.getFirstObject();

        return signerPK.checkSignature(signedData, signature, signatureAlgorithm);

                        // check the signature with appropriate algorithm
//        return signerPK.checkSignature(certificateInfo.toByteArray(), signature, signatureAlgorithm);

    }

    /**
     * returns the PEM format of the certificate
     */
    public byte[] toPEMArray() {
        return certificate.toByteArrayPEM();
    }

    /**
     * returns the DER format of the certificate
     */
    public byte[] toDERArray() {
        return certificate.toByteArrayDER();
    }

    /**
     * saves the certificate
     *
     * @param fileName the file name to use
     */
    public void saveToFile(String fileName) throws IOException {

        if (netscapeFormat)
            createNetscapeCertificate();

        if (saveFormat == ASN1.PEM)
            certificate.saveToFile(fileName, "-----BEGIN CERTIFICATE-----",
                "-----END CERTIFICATE-----");
        else
            certificate.saveToFile(fileName);
    }

    /**
     * converts the certificate to a text output
     */
    public String toString() {
        return certificate.toString();
    }

    /**
     * sets the file format (ASN1.DER, ASN1.PEM) of the cert
     */
    public void setSaveFormat(int format) {
        saveFormat = format;
    }

    /**
     * sets this certificate to netscape format
     */
    public void setNetscapeFormat(boolean nf) {
        netscapeFormat = nf;
        certificate = x509Certificate;
    }

    /**
     * sets the version of this certificate
     */
    public void setVersion(int v) {
        version = v;
    }

    /**
     * sets the serial number of this certificate
     */
    public void setSerialNumber(BigInt sn) {
        serialNumber = sn;
    }

    /**
     * sets the signature algorithm of this certificate
     */
    public void setSignatureAlgorithmID(String said) {
        signatureAlgorithmID = said;
            // todo: cert must be signed again - ensure
    }

    /**
     * sets the issuer of this certificate
     */
    public void setIssuer(Name is) {
        issuer = is;
    }

    /**
     * sets the not before date of this certificate
     */
    public void setNotBefore(Date nb) {
        notBefore = nb;
    }

    /**
     * sets the not after date of this certificate
     */
    public void setNotAfter(Date na) {
        notAfter = na;
    }

    /**
     * sets the subject of this certificate
     */
    public void setSubject(Name su) {
        subject = su;
    }

    /**
     * sets the public key of this certificate
     */
    public void setPublicKey(PublicKey pk) {
        publicKey = pk;
    }

    /**
     * returns the file format (ASN1.DER, ASN1.PEM) of the cert
     */
    public int saveFormat() {
        return saveFormat;
    }

    /**
     * returns true if this cert is in netscape format
     */
    public boolean netscapeFormat() {
        return netscapeFormat;
    }

    /**
     * returns the version of this certificate
     */
    public int version() {
        return version;
    }

    /**
     * returns the serial number of this certificate
     */
    public BigInt serialNumber() {
        return serialNumber;
    }

    /**
     * returns the signature algorithm of this certificate
     */
    public String signatureAlgorithmID() {
        return signatureAlgorithmID;
    }

    /**
     * returns the issuer of this certificate
     */
    public Name issuer() {
        return issuer;
    }

    /**
     * returns the not before date of this certificate
     */
    public Date notBeforeDate() {
        return notBefore;
    }

    /**
     * returns the not after date of this certificate
     */
    public Date notAfterDate() {
        return notAfter;
    }

    /**
     * returns the subject of this certificate
     */
    public Name subject() {
        return subject;
    }

    /**
     * returns the public key of this certificate
     */
    public PublicKey publicKey() {
        return publicKey;
    }

    /**
     * returns the fingerprint of this certificate
     */
    public byte[] fingerprint() {
        return x509Certificate.fingerprint();
    }


    public String toSSLeay() {

        String str = new String();

        str = str.concat(publicKey.toString());
        str = str.concat("\n");
        str = str.concat("Signature:\n");

        byte[] signature = null;

        try {
                        // get the signature from the certificate
            ASN1Object sigObject = x509Certificate.getComponentAt(2);
            signature = (byte[])(sigObject).getValue();
        } catch (CodingException ex) {
            System.out.println("Error!");
        }

        str = str.concat(Util.byteArrayToString(signature));
        return str;
    }
}
