//
// Convert.java --
//
// Author: Pekka Nikander <pnr@tequila.nixu.fi>
//
// Copyright 1996 by Nixu Oy, Helsinki, Finland
//
// Created: Fri Mar 29 20:23:43 1996
// Last modified: Mon Apr  1 00:03:17 1996 pnr
//

package cryptiX.utils;

/**
 * The Convert is a Class only object capable dynamically
 * converting most Java basic types.
 * <P>
 * The purpose of the Convert class is to provide a means to
 * transform a Java object into an object of some other kind,
 * deciding the required conversion type at run time.
 */

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.StringTokenizer;

public class Convert {
  /**
   * Class literals
   */
  private final static Class _String       = new String("").getClass();
  private final static Class _Integer      = new Integer(0).getClass();
  private final static Class _Long         = new Long(0).getClass();
  private final static Class _Float        = new Float(0).getClass();
  private final static Class _Double       = new Double(0).getClass();
  private final static Class _Boolean      = new Boolean(false).getClass();
  private final static Class _Character    = new Character(' ').getClass();
  private final static Class _StringBuffer = new StringBuffer().getClass();
  private final static Class _Object       = new Object().getClass();
  private static Class _InetAddress;
  static {
    try {
      _InetAddress = InetAddress.getLocalHost().getClass();
    } catch (java.net.UnknownHostException e) {
      System.err.println("Java does not support Class literals (InetAddress)");
    }
  }

  /**
   * Restrict creation
   */
  private Convert() {}

  /**
   * Converts the given Java object to another object of the given class.
   * @param object The object to convert
   * @param type   The type of conversion result
   * @return an object representing the same value as <VAR>object</VAR>,
   *         but being of class <VAR>type</VAR>
   */
  public static Object convert(Object object, Class type) {
    // Get the current type of the input object
    Class oldtype = object.getClass();

    // Try if the requested type is an interface the current class implements
    Class[] interfaces = oldtype.getInterfaces();
    for (int i = 0; i < interfaces.length; i++)
      if (interfaces[i].equals(type)) return object;

    // Try superclass relationship, i.e. superclass of current object is
    // requested
    if (oldtype.equals(type)) return object;
    while (!oldtype.equals(_Object)) {
      oldtype = oldtype.getSuperclass();
      if (oldtype.equals(type)) return object;
    }

    // See if we can use the Number interface
    if (object instanceof Number) {
      if (type.equals(_Integer))
	return new Integer(((Number)object).intValue());
      if (type.equals(_Long))
	return new Long(((Number)object).longValue());
      if (type.equals(_Float))
	return new Float(((Number)object).floatValue());
      if (type.equals(_Double))
	return new Double(((Number)object).doubleValue());
    }

    // Ok, convert the value to string and try to convert back...
    String string;

    if (object instanceof String)
      string = (String)object;
    else
      string = object.toString();

    // How about an InetAddress?  (Don't rely on toString)
    if (object instanceof InetAddress) {
      string = ((InetAddress)object).getHostName();
    }

    // See if the user wants to have the string
    if (type.equals(_String))
      return string;

    try {
      if (type.equals(_Integer))      return new Integer(string);
      if (type.equals(_Long))         return new Long(string);
      if (type.equals(_Float))        return new Float(string);
      if (type.equals(_Double))       return new Double(string);
      if (type.equals(_Boolean))      return new Boolean(string);
      if (type.equals(_Character))    return new Character(string.charAt(0));
      if (type.equals(_StringBuffer)) return new StringBuffer(string);
    } catch (Exception e) {
      return null;
    }
    try {
      if (type.equals(_InetAddress))
	return (InetAddress.getAllByName(string))[0];
    } catch (UnknownHostException e) {
      return null;
    }

    // Give up
    return null;
  }
}
